''' MathML Operators Dictionary

    Defines how much space to put around operators such as in "x + y".
    Depends on where the operator falls in the mrow (prefix, infix, or postfix)

    See Appendix F of MathML2 Documentation:
    https://www.w3.org/TR/MathML2/appendixf.html
'''

operators = {
    ('(', 'prefix'):  {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    (')', 'postfix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('[', 'prefix'):  {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    (']', 'postfix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('{', 'prefix'):  {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('}', 'postfix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('”', 'postfix'): {'fence': 'true',  'lspace': '0em', 'rspace': '0em'},  # &CloseCurlyDoubleQuote;
    ('’', 'postfix'): {'fence': 'true',  'lspace': '0em', 'rspace': '0em'},  # &CloseCurlyQuote;
    ('〈', 'prefix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftAngleBracket;
    ('⌈', 'prefix'):  {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftCeiling;
    ('⟪', 'prefix'):  {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftDoubleBracket;
    ('⌊', 'prefix'):  {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftFloor;
    ('“', 'prefix'):  {'fence': 'true',  'lspace': '0em', 'rspace': '0em'},  # &OpenCurlyDoubleQuote;
    ('‘', 'prefix'): {'fence': 'true',  'lspace': '0em', 'rspace': '0em'},  # &OpenCurlyQuote;
    ('〉', 'postfix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &RightAngleBracket;
    ('⌉', 'postfix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &RightCeiling;
    ('〛', 'postfix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &RightDoubleBracket;
    ('⌋', 'postfix'): {'fence': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &RightFloor;
    # ('&InvisibleComma;', 'infix'): {'separator': 'true',  'lspace': '0em', 'rspace': '0em'},
    (',', 'infix'): {'separator': 'true',  'lspace': '0em', 'rspace': 'verythickmathspace'},
    ('─', 'infix'): {'stretchy': 'true', 'minsize':'0',  'lspace': '0em', 'rspace': '0em'},  # '&HorizontalLine;'
    (';', 'infix'): {'separator': 'true',  'lspace': '0em', 'rspace': 'thickmathspace'},
    (';', 'postfix'): {'separator': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('≔', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Assign; or ':='
    ('∵', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Because;
    ('∴', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Therefore;
    ('❘', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &VerticalSeparator;
    ('//', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('∷', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Colon;
    ('&amp;', 'prefix'): {'lspace': '0em', 'rspace': 'thickmathspace'},
    ('&amp;', 'postfix'): {'lspace': 'thickmathspace', 'rspace': '0em'},
    ('⩮', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # *=
    ('-=', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('+=', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('/=', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('->', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    (':', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('..', 'postfix'): {'lspace': 'mediummathspace', 'rspace': '0em'},
    ('...', 'postfix'): {'lspace': 'mediummathspace', 'rspace': '0em'},
    ('∋', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SuchThat;
    ('⫤', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DoubleLeftTee;
    ('⊨', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DoubleRightTee;
    ('⊤', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownTee;
    ('⊣', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftTee;
    ('⊢', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightTee;
    ('⥰', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RoundImplies;
    ('|', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('||', 'infix'): {'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},
    ('⩔', 'infix'): {'stretchy': 'true',  'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &Or;
    ('&&', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('⩓', 'infix'): {'stretchy': 'true',  'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &And;
    ('&', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('!', 'prefix'): {'lspace': '0em', 'rspace': 'thickmathspace'},
    ('⫬', 'prefix'): {'lspace': '0em', 'rspace': 'thickmathspace'},  # &Not;
    ('∃', 'prefix'): {'lspace': '0em', 'rspace': 'thickmathspace'},  # &Exists;
    ('∀', 'prefix'): {'lspace': '0em', 'rspace': 'thickmathspace'},  # &ForAll;
    ('∄', 'prefix'): {'lspace': '0em', 'rspace': 'thickmathspace'},  # &NotExists;
    ('∈', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Element;
    ('∉', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotElement;
    ('∌', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotReverseElement;
    ('⋤', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSquareSubset;
    ('⋢', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'}, # &NotSquareSubsetEqual;
    ('⋥', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'}, # &NotSquareSuperset;
    ('⋣', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSquareSupersetEqual;
    ('⊄', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSubset;
    ('⊈', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSubsetEqual;
    ('⊅', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSuperset;
    ('⊉', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSupersetEqual;
    ('⊏', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SquareSubset;
    ('⊑', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SquareSubsetEqual;
    ('⊐', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SquareSuperset;
    ('⊒', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SquareSupersetEqual;
    ('⋐', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Subset;
    ('⊆', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SubsetEqual;
    ('⊃', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Superset;
    ('⊇', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SupersetEqual;
    ('⇐', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DoubleLeftArrow;
    ('⇔', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DoubleLeftRightArrow;
    ('⇒', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DoubleRightArrow;
    ('⥐', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownLeftRightVector;
    ('⥞', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownLeftTeeVector;
    ('↽', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownLeftVector;
    ('⥖', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownLeftVectorBar;
    ('⥟', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownRightTeeVector;
    ('⇁', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownRightVector;
    ('⥗', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DownRightVectorBar;
    ('←', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftArrow;
    ('⇤', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftArrowBar;
    ('⇆', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftArrowRightArrow;
    ('↔', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftRightArrow;
    ('⥎', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftRightVector;
    ('↤', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftTeeArrow;
    ('⥚', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftTeeVector;
    ('↼', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftVector;
    ('⥒', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftVectorBar;
    ('↙', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LowerLeftArrow;
    ('↘', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LowerRightArrow;
    ('→', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightArrow;
    ('⇥', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightArrowBar;
    ('⇄', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightArrowLeftArrow;
    ('↦', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightTeeArrow;
    ('⥛', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightTeeVector;
    ('⇀', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightVector;
    ('⥓', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightVectorBar;
    ('↖', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &UpperLeftArrow;
    ('↗', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &UpperRightArrow;''
    ('<', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('>', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('≠', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # != or &NotEqual
    ('⩵', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('≤', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},    # &leq 
    ('≥', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},    # & geq
    ('≡', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Congruent;
    ('≍', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &CupCap;
    ('≐', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DotEqual;
    ('∥', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &DoubleVerticalBar;
    ('=', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},   # &Equal;
    ('≂', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &EqualTilde;
    ('⇌', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Equilibrium;
    #('&GreaterEqual;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # Same as &ge;
    ('⋛', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &GreaterEqualLess;
    ('≧', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &GreaterFullEqual;
    ('⪢', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &GreaterGreater;
    ('≷', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &GreaterLess;
    ('⩾', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &GreaterSlantEqual;
    ('≳', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &GreaterTilde;
    ('≎', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &HumpDownHump;
    ('≏', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &HumpEqual;
    ('⊲', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftTriangle;
    ('⧏', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftTriangleBar;
    ('⊴', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LeftTriangleEqual;
    ('⋚', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LessEqualGreater;
    ('≦', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LessFullEqual;
    ('≶', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LessGreater;
    ('⪡', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LessLess;
    ('⩽', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LessSlantEqual;
    ('≲', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &LessTilde;
    ('≫', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NestedGreaterGreater;
    ('≪', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NestedLessLess;
    ('≢', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotCongruent;
    ('≭', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotCupCap;
    ('∦', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotDoubleVerticalBar;
    ('≄', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotEqualTilde;
    ('≯', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotGreater;
    ('≱', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotGreaterEqual;
    ('≩', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotGreaterFullEqual;
    # ('&NotGreaterGreater;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('≹', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotGreaterLess;
    # ('&NotGreaterSlantEqual;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('≵', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotGreaterTilde;
    # ('&NotHumpDownHump;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    # ('&NotHumpEqual;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('⋪', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  #&NotLeftTriangle;
    # ('&NotLeftTriangleBar;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('⋬', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotLeftTriangleEqual;
    ('≮', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotLess;
    ('≰', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotLessEqual;
    ('≸', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotLessGreater;
    # ('&NotLessLess;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ### ??? ('&NotLessSlantEqual;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('≴', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotLessTilde;
    # ('&NotNestedGreaterGreater;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    # ('&NotNestedLessLess;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('⊀', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotPrecedes;
    ('⪵', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotPrecedesEqual;
    ('⋠', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotPrecedesSlantEqual;
    ('⋫', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotRightTriangle;
    # ('&NotRightTriangleBar;', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},
    ('⋭', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotRightTriangleEqual;
    ('⊁', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSucceeds;
    ('⪶', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSucceedsEqual;
    ('⋡', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSucceedsSlantEqual;
    ('⋩', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotSucceedsTilde;
    ('≁', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotTilde;
    ('≇', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotTildeFullEqual;
    ('≉', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotTildeTilde;
    ('∤', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &NotVerticalBar;
    ('≺', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Precedes;
    ('⪯', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &PrecedesEqual;
    ('≼', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &PrecedesSlantEqual;
    ('≾', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &PrecedesTilde;
    ('∝', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Proportional;
    ('⇋', 'infix'): {'stretchy': 'true',  'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &ReverseEquilibrium;
    ('⊳', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightTriangle;
    ('⧐', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightTriangleBar;
    ('⊵', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &RightTriangleEqual;
    ('≻', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Succeeds;
    ('⪰', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SucceedsEqual;
    ('≽', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SucceedsSlantEqual;
    ('≿', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &SucceedsTilde;
    ('∼', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &Tilde;
    ('≃', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &TildeEqual;
    ('≅', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &TildeFullEqual;
    ('≈', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &TildeTilde;
    ('⊥', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &UpTee;
    ('∣', 'infix'): {'lspace': 'thickmathspace', 'rspace': 'thickmathspace'},  # &VerticalBar;
    ('⊔', 'infix'): {'stretchy': 'true',  'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &SquareUnion;
    ('⋃', 'infix'): {'stretchy': 'true',  'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &Union;
    ('⊎', 'infix'): {'stretchy': 'true',  'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &UnionPlus;
    ('−', 'infix'): {'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},
    ('+', 'infix'): {'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},
    ('⋂', 'infix'): {'stretchy': 'true',  'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &Intersection;
    ('∓', 'infix'): {'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &MinusPlus;
    ('±', 'infix'): {'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &PlusMinus;
    ('⊓', 'infix'): {'stretchy': 'true',  'lspace': 'mediummathspace', 'rspace': 'mediummathspace'},  # &SquareIntersection;
    ('⋁', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &Vee;
    ('⊖', 'prefix'): {'largeop': 'true', 'movablelimits': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &CircleMinus;
    ('⊕', 'prefix'): {'largeop': 'true', 'movablelimits': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &CirclePlus;
    ('∑', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &Sum;
    ('⋃', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},
    ('⊎', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},
    ('lim', 'prefix'): {'movablelimits': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},
    ('max', 'prefix'): {'movablelimits': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},
    ('min', 'prefix'): {'movablelimits': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},
    ('⊖', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},
    ('⊕', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},
    ('∫', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &Integral;
    ('∬', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('∭', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('∰', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},
    ('∲', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &ClockwiseContourIntegral;
    ('∮', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &ContourIntegral;
    ('∳', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &CounterClockwiseContourIntegral;
    ('∯', 'prefix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DoubleContourIntegral;
    ('∫', 'infix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &Integral;
    ('∲', 'infix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &ClockwiseContourIntegral;
    ('∮', 'infix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &ContourIntegral;
    ('∳', 'infix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &CounterClockwiseContourIntegral;
    ('∯', 'infix'): {'largeop': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DoubleContourIntegral;
    ('⋓', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Cup;
    ('⋒', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Cap;
    ('≀', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &VerticalTilde;
    ('⋀', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &Wedge;
    ('⊗', 'prefix'): {'largeop': 'true', 'movablelimits': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &CircleTimes;
    ('∐', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &Coproduct;
    ('∏', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &Product;
    ('⋂', 'prefix'): {'largeop': 'true', 'movablelimits': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},
    ('∐', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Coproduct;
    ('⋆', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Star;
    ('⊙', 'prefix'): {'largeop': 'true', 'movablelimits': 'true',  'lspace': '0em', 'rspace': 'thinmathspace'},  # &CircleDot;
    ('*', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},
    ('', 'infix'): {'lspace': '0em', 'rspace': '0em'},  # &InvisibleTimes;
    ('·', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &CenterDot;
    ('⊗', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &CircleTimes;
    ('⋁', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Vee;
    ('⋀', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Wedge;
    ('⋄', 'infix'): {'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Diamond;
    ('∖', 'infix'): {'stretchy': 'true',  'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},  # &Backslash;
    ('/','infix'): {'stretchy': 'true',  'lspace': 'thinmathspace', 'rspace': 'thinmathspace'},
    ('−', 'prefix'): {'lspace': '0em', 'rspace': 'veryverythinmathspace'},
    ('+', 'prefix'): {'lspace': '0em', 'rspace': 'veryverythinmathspace'},
    ('∓', 'prefix'): {'lspace': '0em', 'rspace': 'veryverythinmathspace'},
    ('±', 'prefix'): {'lspace': '0em', 'rspace': 'veryverythinmathspace'},
    ('.', 'infix'): {'lspace': '0em', 'rspace': '0em'},
    ('⨯', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &Cross;
    ('**', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},
    ('⊙', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},
    ('∘', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &SmallCircle;
    ('□', 'prefix'): {'lspace': '0em', 'rspace': 'verythinmathspace'},  # &Square;
    ('∇', 'prefix'): {'lspace': '0em', 'rspace': 'verythinmathspace'},  # &Del;
    ('∂', 'prefix'): {'lspace': '0em', 'rspace': 'verythinmathspace'},  # &PartialD;
    ('ⅅ', 'prefix'): {'lspace': '0em', 'rspace': 'verythinmathspace'},  # &CapitalDifferentialD;
    ('ⅆ', 'prefix'): {'lspace': '0em', 'rspace': 'verythinmathspace'},  # &DifferentialD;
    ('√', 'prefix'): {'stretchy': 'true',  'lspace': '0em', 'rspace': 'verythinmathspace'},  # &Sqrt;
    ('⇓', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DoubleDownArrow;
    ('⟸', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DoubleLongLeftArrow;
    ('⟺', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DoubleLongLeftRightArrow;
    ('⟹', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DoubleLongRightArrow;
    ('⇑', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DoubleUpArrow;
    ('⇕', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DoubleUpDownArrow;
    ('↓', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DownArrow;
    ('⤓', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DownArrowBar;
    ('⇵', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DownArrowUpArrow;
    ('↧', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &DownTeeArrow;
    ('⥡', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LeftDownTeeVector;
    ('⇃', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LeftDownVector;
    ('⥙', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LeftDownVectorBar;
    ('⥑', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LeftUpDownVector;
    ('⥠', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LeftUpTeeVector;
    ('↿', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LeftUpVector;
    ('⥘', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LeftUpVectorBar;
    ('⟵', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LongLeftArrow;
    ('⟷', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LongLeftRightArrow;
    ('⟶', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &LongRightArrow;
    ('⥯', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &ReverseUpEquilibrium;
    ('⥝', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &RightDownTeeVector;
    ('⇂', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &RightDownVector;
    ('⥕', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &RightDownVectorBar;
    ('⥏', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &RightUpDownVector;
    ('⥜', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &RightUpTeeVector;
    ('↾', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &RightUpVector;
    ('⥔', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &RightUpVectorBar;
    ('↑', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &ShortUpArrow;
    ('↑', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &UpArrow;
    ('⤒', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &UpArrowBar;
    ('⇅', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &UpArrowDownArrow;
    ('↕', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &UpDownArrow;
    ('⥮', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &UpEquilibrium;
    ('↥', 'infix'): {'stretchy': 'true',  'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},  # &UpTeeArrow;
    ('^', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},
    ("'", 'postfix'): {'lspace': 'verythinmathspace', 'rspace': '0em'},
    ('!', 'postfix'): {'lspace': 'verythinmathspace', 'rspace': '0em'},
    ('!!', 'postfix'): {'lspace': 'verythinmathspace', 'rspace': '0em'},
    ('~', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},
    ('@', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},
    ('--', 'postfix'): {'lspace': 'verythinmathspace', 'rspace': '0em'},
    ('--', 'prefix'): {'lspace': '0em', 'rspace': 'verythinmathspace'},
    ('++', 'postfix'): {'lspace': 'verythinmathspace', 'rspace': '0em'},
    ('++', 'prefix'): {'lspace': '0em', 'rspace': 'verythinmathspace'},
    (' ', 'infix'): {'lspace': '0em', 'rspace': '0em'},  # &ApplyFunction;
    ('?', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},
    ('_', 'infix'): {'lspace': 'verythinmathspace', 'rspace': 'verythinmathspace'},
    ('˘', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &Breve;
    ('¸', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &Cedilla;
    ('`', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DiacriticalGrave;
    ('˙', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DiacriticalDot;
    ('˝', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DiacriticalDoubleAcute;
    ('←', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftArrow;
    ('↔', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftRightArrow;
    ('⥎', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftRightVector;
    ('↼', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &LeftVector;
    ('´', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DiacriticalAcute;
    ('→', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &RightArrow;
    ('⇀', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  #&RightVector;
    ('˜', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DiacriticalTilde;
    ('¨', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &DoubleDot;
    ('ˇ', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &Hacek;
    ('̂', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &Hat;
    ('‾', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &OverBar;
    ('⏞', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &OverBrace;
    ('⎴', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &OverBracket;
    ('⏜', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  #&OverParenthesis;
    ('…', 'postfix'): {'accent': 'true',  'lspace': '0em', 'rspace': '0em'},  # &TripleDot;
    ('_', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &UnderBar;
    ('⏟', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &UnderBrace;
    ('⎵', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &UnderBracket;
    ('⏝', 'postfix'): {'accent': 'true', 'stretchy': 'true',  'lspace': '0em', 'rspace': '0em'},  # &UnderParenthesis;
}


integrals = ['∫', '∬', '∭', '∲', '∮', '∳', '∯', '∰', ]


''' Apply italic, bold, and other font styles by shifting the unstyled ASCII
    characters [A-Z, a-z, and 0-9] to their higher unicode alternatives. Note
    this does not check whether the new character glyph exists in the font.
'''

from __future__ import annotations
from typing import Literal

StyleType = Literal['serif', 'sans', 'script', 'fraktur', 'double', 'mono']


def styledchr(c: str, italic: bool=False, bold: bool=False,
              style: StyleType='serif') -> str:
    ''' Convert an ASCII character into it's styled unicode counterpart '''
    ordc = ord(c)
    if not ((0x41 <= ordc <= 0x5A) or (0x61 <= ordc <= 0x7A) or (0x30 <= ordc <= 0x39)):
        return c

    # Look up replacement characters that don't map directly by offset
    # because unicode defines them in a weird place.
    replace = {('serif', True, False): {
                   'h': 'ℎ'},
               ('script', False, False): {
                   'B': 'ℬ', 'E': 'ℰ', 'F': 'ℱ', 'H': 'ℋ', 'I': 'ℐ', 'L': 'ℒ',
                   'M': 'ℳ', 'R': 'ℛ', 'e': 'ℯ', 'g': 'ℊ', 'o': 'ℴ'},
                ('fraktur', False, False): {
                    'C': 'ℭ', 'H': 'ℌ', 'I': 'ℑ', 'R': 'ℜ', 'Z': 'ℨ'},
                ('double', False, False): {
                    'C': 'ℂ', 'H': 'ℍ', 'N': 'ℕ', 'P': 'ℙ', 'Q': 'ℚ', 'R': 'ℝ', 'Z': 'ℤ'}}

    cx = replace.get((style, italic, bold), {}).get(c)
    if cx is not None:
        return cx

    if (0x30 <= ordc <= 0x39):  # Digit
        #        style    bold
        ofst = {('serif', True): 0x1D7CE,
                ('double', False): 0x1D7D8,
                ('sans', False): 0x1D7E2,
                ('sans', True): 0x1D7EC,
                ('mono', False): 0x1D7F6}.get((style, bold))
        if ofst is None:
            return c
        return chr(ordc - ord('0') +ofst)

    if (0x61 <= ordc <= 0x7A):
        # Lowercase
        base = ordc - 6 - ord('A')
    else:
        base = ordc - ord('A')

    ofst = None
    if style == 'serif':
        ofst = {(False, True): 0x1D400,
                (True, False): 0x1D434,
                (True, True): 0x1D468}.get((italic, bold))
    elif style == 'sans':
        ofst = {(False, False): 0x1D5A0,
                (False, True): 0x1D5D4,
                (True, False): 0x1D608,
                (True, True): 0x1D63C}.get((italic, bold))
    elif style == 'script':
        ofst = {False: 0x1D49C,
                True: 0x1D4D0}.get(bold)
    elif style == 'fraktur':
        ofst = {False: 0x1D504,
                True: 0x1D504}.get(bold)
    elif style == 'mono':
        ofst = 0x1D670
    elif style == 'double':
        ofst = 0x1D538

    if ofst is None:
        return c

    return chr(base + ofst)


def styledstr(st: str, italic: bool=False, bold: bool=False, style: StyleType='serif'):
    ''' Apply unicode styling conversion to a string '''
    return ''.join([styledchr(s, italic, bold, style) for s in st])



''' Main math rendering class '''

from __future__ import annotations
from typing import Union, Literal
import re
from itertools import zip_longest
import importlib.resources as pkg_resources
import xml.etree.ElementTree as ET

import ziafont as zf

from .mathfont import MathFont
from .nodes import makenode, getstyle
from .mathtable import MathTable
from .escapes import unescape

try:
    from latex2mathml.converter import convert  # type: ignore
except ImportError:
    convert = False


Halign = Literal['left', 'center', 'right']
Valign = Literal['top', 'center', 'baseline', 'axis', 'bottom']


def denamespace(element: ET.Element) -> ET.Element:
    ''' Recursively remove namespace {...} from beginning of xml
        element names, so they can be searched easily.
    '''
    if element.tag.startswith('{'):
        element.tag = element.tag.split('}')[1]
    for elm in element:
        denamespace(elm)
    return element


class Math:
    ''' Math Renderer

        Args:
            mathml: MathML expression, in string or XML Element
            size: Base font size, pixels
            font: Filename of font file. Must contain MATH typesetting table.
    '''
    def __init__(self, mathml: Union[str, ET.Element],
                 size: float=24, font: str=None):
        self.mathml = mathml
        self.size = size

        if font is None:
            self.font = loadedfonts.get('default')
        elif font in loadedfonts:
            self.font = loadedfonts.get(font)
        else:
            self.font = MathFont(font, size)
            loadedfonts[font] = self.font

        if isinstance(mathml, str):
            mathml = unescape(mathml)
            mathml = ET.fromstring(mathml)
        mathml = denamespace(mathml)

        self.style = getstyle(mathml)
        self.element = mathml
        self.node = makenode(mathml, parent=self, size=size)  # type: ignore

    @classmethod
    def fromlatex(cls, latex: str, size: float=24, mathstyle: str=None, font: str=None):
        ''' Create Math Renderer from a single LaTeX expression. Requires
            latex2mathml Python package.

            Args:
                latex: Latex string
                size: Base font size
                mathstyle: Style parameter for math, equivalent to "mathvariant" MathML attribute
                font: Font file name
        '''
        if not convert:
            raise ValueError('fromlatex requires latex2mathml package.')
        mathml = convert(latex)
        if mathstyle:
            mathml = ET.fromstring(mathml)
            mathml.attrib['mathvariant'] = mathstyle
        return cls(mathml, size, font)

    @classmethod
    def fromlatextext(cls, latex: str, size: float=24, mathstyle: str=None,
                      textstyle: str=None, font: str=None):
        ''' Create Math Renderer from a sentance containing zero or more LaTeX
            expressions delimited by $..$.
            Requires latex2mathml Python package.

            Args:
                latex: string
                size: Base font size
                mathstyle: Style parameter for math, equivalent to "mathvariant" MathML attribute
                textstyle: Style parameter for text, equivalent to "mathvariant" MathML attribute
                font: Font file name
        '''
        # Extract each $..$, convert to MathML, but the raw text in <mtext>, and join
        # into a single <math>
        parts = re.split('\$(.*?)\$', latex)
        texts = parts[::2]
        maths = [convert(p) for p in parts[1::2]]
        mathels = [ET.fromstring(m)[0] for m in maths]   # Convert to xml, but drop opening <math>
        mml = ET.Element('math')
        for text, mathel in zip_longest(texts, mathels):
            if text:
                mtext = ET.SubElement(mml, 'mtext')
                if textstyle:
                    mtext.attrib['mathvariant'] = textstyle
                mtext.text = text
            if mathel is not None:
                if mathstyle:
                    mathel.attrib['mathvariant'] = mathstyle
                mml.append(mathel)
        return cls(mml, size, font)
    
    def svgxml(self) -> ET.Element:
        ''' Get standalone SVG of expression as XML Element Tree '''
        svg = ET.Element('svg')
        self.node.draw(1, 0, svg)
        bbox = self.node.bbox
        width = bbox.xmax - bbox.xmin + 2  # Add a 1-px border
        height = bbox.ymax - bbox.ymin + 2

        # Note: viewbox goes negative.
        svg.attrib['width'] = str(width)
        svg.attrib['height'] = str(height)
        svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
        svg.attrib['viewBox'] = f'0 {-bbox.ymax-1} {width} {height}'
        return svg

    def drawon(self, x: float, y: float, svg: ET.Element,
               color: str=None,
               halign: Halign='left', valign: Valign='baseline') -> ET.Element:
        ''' Draw the math expression on an existing SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: The image (XML object) to draw on
                color: Color name or #000000 hex code
                halign: Horizontal alignment
                valign: Vertical alignment

            Note: Horizontal alignment can be the typical 'left', 'center', or 'right'.
            Vertical alignment can be 'top', 'bottom', or 'center' to align with the
            expression's bounding box, or 'baseline' to align with the bottom
            of the first text element, or 'axis', aligning with the height of a minus
            sign above the baseline.
        '''
        width, height = self.getsize()
        yshift = {'top': self.node.bbox.ymax,
                  'center': height/2 + self.node.bbox.ymin,
                  'axis': self.font.math.consts.axisHeight * self.node.emscale,  # type: ignore
                  'bottom': self.node.bbox.ymin}.get(valign, 0)
        xshift = {'center': -width/2,
                  'right': -width}.get(halign, 0)

        svgelm = ET.SubElement(svg, 'g')  # Put it in a group
        if color:
            svgelm.attrib['fill'] = color
        self.node.draw(x+xshift, y+yshift, svgelm)
        return svgelm

    def svg(self) -> str:
        ''' Get expression as SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def _repr_svg_(self):
        ''' Jupyter SVG representation '''
        return self.svg()

    @classmethod
    def mathml2svg(cls, mathml: Union[str, ET.Element],
                   size: float=24, font: str=None):
        ''' Shortcut to just return SVG string directly '''
        return cls(mathml, size=size, font=font).svg()

    def getsize(self) -> tuple[float, float]:
        ''' Get size of rendered text '''
        return (self.node.bbox.xmax - self.node.bbox.xmin,
                self.node.bbox.ymax - self.node.bbox.ymin)
    
    
# Cache the loaded fonts to prevent reloading all the time
with pkg_resources.path('ziamath.fonts', 'STIXTwoMath-Regular.ttf') as p:
    fontname = p
loadedfonts = {'default': MathFont(fontname)}



''' Math Font extends Ziafont with MATH table '''

from __future__ import annotations
from typing import Union
from pathlib import Path

from ziafont import Font

from .mathtable import MathTable


class MathFont(Font):
    ''' Extend ziafont by reading MATH table and a base font size

        Args:
            fname: File name of font
            basesize: Default font size
    '''
    def __init__(self, fname: Union[str, Path], basesize: float=24):
        super().__init__(fname)
        self.basesize = basesize
        if 'MATH' not in self.tables:
            raise ValueError('Font has no MATH table!')
        self.math = MathTable(self)




''' Math element nodes '''

from __future__ import annotations
from typing import Optional, Union, MutableMapping

from copy import copy
from collections import ChainMap
import xml.etree.ElementTree as ET

from ziafont import Font
from ziafont.fonttypes import BBox
from ziafont.glyph import SimpleGlyph

from .styles import styledstr
from .zmath import MathFont
from . import operators


DEBUG = False       # Debug mode, draws bboxes
MINFONTSIZE = 0.3   # Minimum font size as fraction of base font size


def getstyle(element: ET.Element) -> dict:
    ''' Get style arguments based on "mathvariant" in mathml '''
    variant = element.attrib.get('mathvariant', '')
    styleargs: dict[str, Union[str, bool]] = {}
    if 'italic' in variant:
        styleargs['italic'] = True
    if 'bold' in variant:
        styleargs['bold'] = True
    if 'double' in variant:
        styleargs['style'] = 'double'
    if 'script' in variant:
        styleargs['style'] = 'script'
    if 'sans' in variant:
        styleargs['style'] = 'sans'
    if 'mono' in variant:
        styleargs['style'] = 'mono'
    if 'fraktur' in variant:
        styleargs['style'] = 'fraktur'
    if 'displaystyle' in element.attrib:
        styleargs['displaystyle'] = element.attrib['displaystyle']
    return styleargs


def makenode(element: ET.Element, size: float,
             parent: 'Mnode', **kwargs) -> 'Mnode':
    ''' Create node from the MathML element

        Args:
            element: MathML XML element
            size: Font size for element
            parent: Parent node
    '''
    node = {'math': Mrow,
            'mrow': Mrow,
            'mi': Midentifier,
            'mn': Mnumber,
            'mo': Moperator,
            'msup': Msup,
            'msub': Msub,
            'msubsup': Msubsup,
            'mover': Mover,
            'munder': Munder,
            'munderover': Munderover,
            'mfrac': Mfrac,
            'msqrt': Msqrt,
            'mroot': Mroot,
            'mtext': Mtext,
            'mspace': Mspace,
            'mpadded': Mpadded,
            'mphantom': Mphantom,
            'mtable': Mtable,
            'mtd': Mrow,
            }.get(element.tag, None)
    if element.tag == 'mo':
        infer_opform(0, element, parent)

    if node:
        return node(element, size, parent, **kwargs)
    else:
        print('Undefined Element', element)
        return Mrow(element, size, parent)


def getspaceems(space: str) -> float:
    ''' Get space in ems from the string. Can be number or named space width. '''
    if space.endswith('em'):
        f = float(space[:-2])
    else:
        f = {"veryverythinmathspace": 1/18,
             "verythinmathspace": 2/18,
             "thinmathspace": 3/18,
             "mediummathspace": 4/18,
             "thickmathspace": 5/18,
             "verythickmathspace": 6/18,
             "veryverythickmathspace": 7/18}.get(space, 0)
    return f


def getdimension(size: str, emscale: float) -> float:
    ''' Get dimension from string. Either in em or px '''
    try:
        s = float(size)
    except ValueError:
        if size.endswith('em'):
            s = float(size[:-2]) / emscale
        elif size.endswith('px'):
            s = float(size[:-2])
        else:
            raise ValueError(f'Undefined size {s}')
    return s


def getelementtext(element: ET.Element) -> str:
    ''' Get text of XML element '''
    try:
        txt = element.text.strip()  # type: ignore
    except AttributeError:
        txt = ''
    return txt


class Mnode:
    ''' Math Drawing Node

        Args:
            element: XML element for the node
            size: font size
            parent: Mnode of parent
    '''
    def __init__(self, element: ET.Element, size: float, parent: 'Mnode', **kwargs):
        self.element = element
        self.font: MathFont = parent.font
        self.size = size
        self.parent = parent
        self.style: MutableMapping[str, Union[str, bool]] = ChainMap(copy(parent.style), getstyle(self.element))
        self.emscale = size / self.font.info.layout.unitsperem
        self.nodes: list[Mnode] = []
        self.nodexy: list[tuple[float, float]] = []

    def _setup(self, **kwargs) -> None:
        ''' Calculate node position assuming this node is at 0, 0. Also set bbox. '''
        self.bbox = BBox(0, 0, 0, 0)

    def leftsibling(self) -> Optional['Mnode']:
        ''' Left node sibling. The one that was just placed. '''
        try:
            return self.parent.nodes[-1]
        except (IndexError, AttributeError):
            return None

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return None

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return None

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        return None

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if DEBUG:
            rect = ET.SubElement(svg, 'rect')
            rect.attrib['x'] = str(x)
            rect.attrib['y'] = str(y - self.bbox.ymax)
            rect.attrib['width'] = str((self.bbox.xmax - self.bbox.xmin))
            rect.attrib['height'] = str((self.bbox.ymax - self.bbox.ymin))
            rect.attrib['fill'] = 'none'
            rect.attrib['stroke'] = 'blue'
            rect.attrib['stroke-width'] = '0.2'

        xi = yi = 0.
        for (xi, yi), node in zip(self.nodexy, self.nodes):
            node.draw(x+xi, y+yi, svg)
        return x+xi, y+yi


class MGlyph(Mnode):
    ''' A single glyph node (not <mglyph>, but used by other nodes) '''
    def __init__(self, glyph: SimpleGlyph, char: str, size: float, emscale: float, **kwargs):
        self.glyph = glyph
        self.char = char
        self.size = size
        self.emscale = emscale
        self.phantom = kwargs.get('phantom', False)
        self._setup()

    def _setup(self, **kwargs) -> None:
        ''' Place the glyphs with 0, 0 positions '''
        self.bbox = BBox(
            self.glyph.path.bbox.xmin * self.emscale,
            self.glyph.path.bbox.xmax * self.emscale,
            self.glyph.path.bbox.ymin * self.emscale,
            self.glyph.path.bbox.ymax * self.emscale)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return self.glyph

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.glyph

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        return self.char

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        symbols = svg.findall('symbol')
        symids = [sym.attrib.get('id') for sym in symbols]
        if self.glyph.id not in symids:
            svg.append(self.glyph.svgsymbol())
        if not self.phantom:
            svg.append(self.glyph.place(x, y, self.size))
        x += self.glyph.advance() * self.emscale
        return x, y


class MHLine(Mnode):
    ''' SVG Horizontal Line. Used by mfrac and msqrt. '''
    def __init__(self, length: float, lw: float, **kwargs):
        self.length = length
        self.lw = lw
        self.phantom = kwargs.get('phantom', False)
        self.bbox = BBox(0, lw/2, self.length, self.lw)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return None

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return None

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        return None

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if not self.phantom:
            # Use rectangle so it can change color with 'fill' attribute
            # and not mess up glyphs with 'stroke' attribute
            bar = ET.SubElement(svg, 'rect')
            bar.attrib['x'] = str(x)
            bar.attrib['y'] = str(y)
            bar.attrib['width'] = str(self.length)
            bar.attrib['height'] = str(self.lw)
        return x+self.length, y


class Midentifier(Mnode):
    ''' Identifier node <mi> '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)

        # Identifiers are italic unless longer than one character
        text = getelementtext(self.element)
        if len(text) == 1 and 'italic' not in self.style:
            self.style['italic'] = True
        if len(text) > 1:
            text += ' '
        self.string = styledstr(text, **self.style)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        ymin = 9999
        ymax = -9999
        x = 0
        for char in self.string:
            glyph = self.font.glyph(char)
            self.nodes.append(MGlyph(glyph, char, self.size, self.emscale, **kwargs))
            self.nodexy.append((x, 0))
            x += glyph.advance() * self.emscale
            ymin = min(ymin, glyph.path.bbox.ymin * self.emscale)
            ymax = max(ymax, glyph.path.bbox.ymax * self.emscale)
        self.bbox = BBox(0, x, ymin, ymax)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        try:
            return self.nodes[0].firstglyph()
        except IndexError:
            return None

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        try:
            return self.nodes[-1].lastglyph()
        except IndexError:
            return None

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        try:
            return self.nodes[-1].lastchar()
        except IndexError:
            return None


class Mnumber(Midentifier):
    ''' Number node <mn> '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        Mnode.__init__(self, element, size, parent, **kwargs)
        self.string = styledstr(getelementtext(self.element), **self.style)
        self._setup(**kwargs)


class Mtext(Midentifier):
    ''' Text Node <mtext> '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        Mnode.__init__(self, element, size, parent, **kwargs)
        # Don't use getelementtext since it strips whitespace
        self.string = ''
        if self.element.text:
            self.string = styledstr(self.element.text, **self.style)
        self._setup(**kwargs)


def infer_opform(i: int, child: ET.Element, mrow: Mnode) -> None:
    ''' Infer form (prefix, postfix, infix) of operator child within
        element mrow. Appends 'form' attribute to child.

        Args:
            i: Index of child within parent
            child: XML element of child
            mrow: Mnode of parent element
    '''
    # Infer form for operators
    if 'form' not in child.attrib:
        if isinstance(mrow, (Msub, Msub, Msubsup)):
            form = 'prefix'
        elif i == 0:
            form = 'prefix'
        elif i == len(mrow.element) - 1:
            form = 'postfix'
        else:
            form = 'infix'
        child.attrib['form'] = form


def isstretchy(text: str, font: MathFont) -> bool:
    ''' Check if glyph is in font's extendedShapeCoverage list '''
    if text:
        glyph = font.glyph(text[0])
        return font.math.isextended(glyph.index)
    return False


class Mrow(Mnode):
    ''' Math row, list of vertically aligned Mnodes '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        kwargs = copy(kwargs)
        for i, child in enumerate(self.element):
            if child.tag == 'mo':
                infer_opform(i, child, self)

        self.nodes = []
        x = 0
        ymax = -9999
        ymin = 9999
        i = 0
        node: Mnode
        while i < len(self.element):
            child = self.element[i]
            text = getelementtext(child)
            if child.tag == 'mo':
                if (isstretchy(text, self.font) and
                    child.attrib.get('form') == 'prefix' and
                    child.attrib.get('stretchy') != 'false'):
                    fencekwargs = copy(kwargs)
                    j = 0
                    for j in range(i+1, len(self.element)):
                        if self.element[j].tag == 'mo' and self.element[j].attrib.get('form') == 'postfix':
                            children = self.element[i+1: j]
                            fencekwargs['open'] = getelementtext(child)
                            fencekwargs['close'] = getelementtext(self.element[j])
                            break
                    else:  # No postfix closing fence. Enclose remainder of row.
                        children = self.element[i+1:]
                        fencekwargs['open'] = getelementtext(child)
                        fencekwargs['close'] = None
                    fenced = ET.Element('fenceop')
                    fenced.attrib.update(child.attrib)
                    frow = ET.SubElement(fenced, 'mrow')
                    frow.extend(children)
                    node = Mopfence(fenced, self.size, parent=self, **fencekwargs)
                    i = j + 1

                else:
                    if text == '':
                        i += 1
                        continue  # InvisibleTimes, etc.
                    node = Moperator(child, self.size, parent=self, **kwargs)
                    i += 1
            else:
                node = makenode(child, self.size, parent=self, **kwargs)
                i += 1
            self.nodes.append(node)
            self.nodexy.append((x, 0))
            x += node.bbox.xmax
            ymax = max(ymax, node.bbox.ymax)
            ymin = min(ymin, node.bbox.ymin)
        self.bbox = BBox(0, x, ymin, ymax)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return self.nodes[0].firstglyph()

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.nodes[-1].lastglyph()

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        try:
            return self.nodes[-1].lastchar()
        except IndexError:
            return None


class Mopfence(Mnode):
    ''' Fenced operator. MathML doesn't put items inside fence () in xml subelement
        which means it's not easy to calculate the stretch size of the fence glyphs.
        This class wraps the content of the fence and handles both opening and closing
        operators.
    '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        assert len(element) == 1  # Should contain 1 mrow with contents of Opfence
        self.openchr = kwargs.get('open')
        self.closechr = kwargs.get('close')
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        mrow = Mrow(self.element[0], self.size, parent=self)
        height = mrow.bbox.ymax - mrow.bbox.ymin
        self.nodes = []
        x = 0
        if height > self.size * 1.5:
            # Tall enclosures (like tables) should be centered vertically
            rowbaseline = -self.font.math.consts.axisHeight * self.emscale
        else:
            rowbaseline = 0
        rowcenter = rowbaseline - mrow.bbox.ymin - (mrow.bbox.ymax - mrow.bbox.ymin)/2

        if self.openchr:
            openglyph = self.font.glyph(self.openchr)
            openglyph = self.font.math.variant(openglyph.index, height/self.emscale, vert=True)
            mglyph = MGlyph(openglyph, self.openchr, self.size, self.emscale, **kwargs)
            yofst = rowcenter + mglyph.bbox.ymin + (mglyph.bbox.ymax - mglyph.bbox.ymin)/2

            self.nodes.append(mglyph)
            self.nodexy.append((x, yofst))
            x += openglyph.advance() * self.emscale

        self.nodes.append(mrow)
        self.nodexy.append((x, rowbaseline))
        x += mrow.bbox.xmax

        if self.closechr:
            closeglyph = self.font.glyph(self.closechr)
            closeglyph = self.font.math.variant(closeglyph.index, height/self.emscale, vert=True)

            yofst = rowcenter + mglyph.bbox.ymin + (mglyph.bbox.ymax - mglyph.bbox.ymin)/2
            self.nodes.append(MGlyph(closeglyph, self.closechr, self.size, self.emscale, **kwargs))
            self.nodexy.append((x, yofst))
            x += closeglyph.advance() * self.emscale

        self.bbox = BBox(0, x, -rowbaseline+mrow.bbox.ymin, -rowbaseline+mrow.bbox.ymax)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return self.nodes[0].firstglyph()

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.nodes[-1].firstglyph()

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        try:
            return self.nodes[-1].lastchar()
        except IndexError:
            return None


class Moperator(Mnumber):
    ''' Operator math element '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        Mnode.__init__(self, element, size, parent, **kwargs)
        self.string = styledstr(getelementtext(self.element), **self.style)
        self.form = element.attrib.get('form', 'infix')

        # Load parameters from operators table for deciding how much space
        # to add on either side of the operator
        self.params = operators.operators.get((self.string, self.form), {})
        self.width = kwargs.get('width', None)
        self._setup(**kwargs)

    def _setup(self, **kwargs):
        if kwargs.get('sup') or kwargs.get('sub'):
            addspace = False  # Dont add lspace/rspace when in super/subscripts
        else:
            addspace = True

        x = 0
        self.nodes = []

        # Add lspace
        if addspace:
            lspace = getspaceems(self.params.get('lspace', '0')) * self.emscale * self.font.info.layout.unitsperem
            x += lspace

        glyphs = [self.font.glyph(char) for char in self.string]
        ymin = 999
        ymax = -999
        for glyph, char in zip(glyphs, self.string):
            if self.params.get('largeop') == 'true' and self.style.get('displaystyle', 'true') != 'false':
                minh = self.font.math.consts.displayOperatorMinHeight
                glyph = self.font.math.variant(glyph.index, minh, vert=True)

            if self.width:
                glyph = self.font.math.variant(glyph.index, self.width / self.emscale, vert=False)

            self.nodes.append(MGlyph(glyph, char, self.size, self.emscale, **kwargs))
            self.nodexy.append((x, 0))
            x += glyph.advance() * self.emscale
            ymin = min(ymin, glyph.path.bbox.ymin * self.emscale)
            ymax = max(ymax, glyph.path.bbox.ymax * self.emscale)

        if addspace:
            rspace = getspaceems(self.params.get('rspace', '0')) * self.emscale * self.font.info.layout.unitsperem
            x += rspace

        self.bbox = BBox(glyphs[0].path.bbox.xmin * self.emscale, x, ymin, ymax)


def place_super(base: Mnode, superscript: Mnode, font: MathFont, emscale: float) -> tuple[float, float, float]:
    ''' Superscript. Can be above the operator (like sum) or regular super '''
    if (hasattr(base, 'params') and base.params.get('movablelimits') == 'true'  # type: ignore
        and base.style.get('displaystyle', 'true') == 'true'):
        x = -(base.bbox.xmax - base.bbox.xmin) / 2 - (superscript.bbox.xmax - superscript.bbox.xmin) / 2
        supy = -base.bbox.ymax - font.math.consts.upperLimitGapMin * emscale + superscript.bbox.ymin
        xadvance = 0
    else:
        x = 0
        lastg = base.lastglyph()
        shiftup = font.math.consts.superscriptShiftUp * emscale

        if hasattr(base, 'params'):
            x -= getspaceems(base.params.get('rspace', '0')) / emscale  # type: ignore

        if lastg:
            italicx = font.math.italicsCorrection.getvalue(lastg.index)
            if italicx and base.lastchar() not in operators.integrals:
                x += italicx * emscale
            firstg = superscript.firstglyph()
            if firstg:
                kern, shiftup = font.math.kernsuper(lastg, firstg)
                x += kern * emscale
            else:
                shiftup -= superscript.bbox.ymin / emscale
        supy = -shiftup * emscale
        xadvance = x + superscript.bbox.xmax
    return x, supy, xadvance


def place_sub(base: Mnode, subscript: Mnode, font: MathFont, emscale: float):
    ''' Calculate subscript. Can be below the operator (like sum) or regular sub '''
    if hasattr(base, 'params') and base.params.get('movablelimits') == 'true' and base.style.get('displaystyle', 'true') == 'true':  # type: ignore
        x = -(base.bbox.xmax - base.bbox.xmin) / 2 - (subscript.bbox.xmax - subscript.bbox.xmin) / 2
        suby = -base.bbox.ymin + font.math.consts.lowerLimitGapMin * emscale + subscript.bbox.ymax
        xadvance = 0
    else:
        lastg = base.lastglyph()
        shiftdn = font.math.consts.subscriptShiftDown
        x = 0
        if hasattr(base, 'params'):
            x -= getspaceems(base.params.get('rspace', '0')) / emscale  # type: ignore

        if lastg:
            italicx = font.math.italicsCorrection.getvalue(lastg.index)
            if italicx and base.lastchar() in operators.integrals:
                x -= italicx * emscale  # Shift back on integrals
            firstg = subscript.firstglyph()
            if firstg:
                kern, shiftdn = font.math.kernsub(lastg, firstg)
                x += kern * emscale
            else:
                shiftdn -= subscript.bbox.ymin / emscale
        suby = shiftdn * emscale
        xadvance = x + subscript.bbox.xmax
    return x, suby, xadvance


class Msup(Mnode):
    ''' Superscript Node '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        assert len(self.element) == 2
        self.base = makenode(self.element[0], size, parent=self, **kwargs)
        kwargs['sup'] = True
        supfontsize = max(self.size * self.font.math.consts.scriptPercentScaleDown / 100,
                          self.font.basesize*MINFONTSIZE)
        self.superscript = makenode(self.element[1], supfontsize, parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes.append(self.base)
        self.nodexy.append((0, 0))
        x = self.base.bbox.xmax

        supx, supy, xadv = place_super(self.base, self.superscript, self.font, self.emscale)
        self.nodes.append(self.superscript)
        self.nodexy.append((x+supx, supy))
        xmin = self.base.bbox.xmin
        xmax = x + xadv
        ymin = self.base.bbox.ymin
        ymax = max(self.base.bbox.ymax, -supy + self.superscript.bbox.ymax)
        self.bbox = BBox(xmin, xmax, ymin, ymax)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return self.nodes[0].firstglyph()

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.nodes[-1].lastglyph()

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        try:
            return self.nodes[-1].lastchar()
        except IndexError:
            return None


class Msub(Mnode):
    ''' Subscript Node '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        assert len(self.element) == 2
        self.base = makenode(self.element[0], size, parent=self, **kwargs)
        kwargs['sub'] = True
        subfontsize = max(self.size * self.font.math.consts.scriptPercentScaleDown / 100,
                          self.font.basesize*MINFONTSIZE)
        self.subscript = makenode(self.element[1], subfontsize, parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes.append(self.base)
        self.nodexy.append((0, 0))
        x = self.base.bbox.xmax

        subx, suby, xadv = place_sub(self.base, self.subscript, self.font, self.emscale)
        self.nodes.append(self.subscript)
        self.nodexy.append((x + subx, suby))
        xmin = self.base.bbox.xmin
        xmax = x + xadv
        ymin = min(self.base.bbox.ymin, -suby+self.subscript.bbox.ymin)
        ymax = max(self.base.bbox.ymax, -suby+self.subscript.bbox.ymax)
        self.bbox = BBox(xmin, xmax, ymin, ymax)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return self.nodes[0].firstglyph()

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.nodes[-1].lastglyph()

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        try:
            return self.nodes[-1].lastchar()
        except IndexError:
            return None


class Msubsup(Mnode):
    ''' Subscript and Superscript together '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        assert len(self.element) == 3
        self.base = makenode(self.element[0], size, parent=self, **kwargs)
        subfontsize =  max(self.size * self.font.math.consts.scriptPercentScaleDown / 100,
                           self.font.basesize*MINFONTSIZE)
        kwargs['sup'] = True
        self.subscript = makenode(self.element[1], subfontsize, parent=self, **kwargs)
        kwargs['sub'] = True
        self.superscript = makenode(self.element[2], subfontsize, parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes.append(self.base)
        self.nodexy.append((0, 0))
        x = self.base.bbox.xmax
        subx, suby, xadvsub = place_sub(self.base, self.subscript, self.font, self.emscale)
        supx, supy, xadvsup = place_super(self.base, self.superscript, self.font, self.emscale)

        # Ensure subSuperscriptGapMin between scripts
        if (suby - self.subscript.bbox.ymax) - (supy-self.superscript.bbox.ymin) < self.font.math.consts.subSuperscriptGapMin*self.emscale:
            diff = self.font.math.consts.subSuperscriptGapMin*self.emscale - (suby - self.subscript.bbox.ymax) + (supy-self.superscript.bbox.ymin)
            suby += diff/2
            supy -= diff/2

        self.nodes.append(self.subscript)
        self.nodexy.append((x + subx, suby))
        self.nodes.append(self.superscript)
        self.nodexy.append((x + supx, supy))

        xmin = self.base.bbox.xmin
        xmax = max(x + xadvsup, x + xadvsub)
        ymin = min(-self.base.bbox.ymin, -suby+self.subscript.bbox.ymin)
        ymax = max(self.base.bbox.ymax, -supy + self.superscript.bbox.ymax)
        self.bbox = BBox(xmin, xmax, ymin, ymax)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return self.nodes[0].firstglyph()

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.nodes[-1].lastglyph()

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        try:
            return self.nodes[-1].lastchar()
        except IndexError:
            return None


def place_over(base: Mnode, over: Mnode, font: MathFont, emscale: float) -> tuple[float, float]:
    ''' Place node over another one

        Args:
            base: Base node
            over: Node to draw over the base
            font: Font
            emscale: Font scale

        Returns:
            x, y: position for over node
    '''
    # TODO accent parameter used to raise/lower
    x = ((base.bbox.xmax - base.bbox.xmin) - (over.bbox.xmax-over.bbox.xmin)) / 2
    y = -base.bbox.ymax-font.math.consts.overbarVerticalGap*emscale
    y += over.bbox.ymin
    return x, y


class Mover(Mnode):
    ''' Over bar node '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        kwargs = copy(kwargs)
        assert len(self.element) == 2
        self.base = makenode(self.element[0], size, parent=self, **kwargs)
        kwargs['sup'] = True
        fsize = max(self.size * self.font.math.consts.scriptPercentScaleDown / 100,
                          self.font.basesize*MINFONTSIZE)

        kwargs['width'] = self.base.bbox.xmax - self.base.bbox.xmin
        self.over = makenode(self.element[1], fsize, parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        overx, overy = place_over(self.base, self.over, self.font, self.emscale)

        basex = 0.
        if overx < 0:
            basex = -overx
            overx = 0.

        self.nodes.append(self.base)
        self.nodexy.append((basex, 0))
        self.nodes.append(self.over)
        self.nodexy.append((overx, overy))
        xmin = min(overx, self.base.bbox.xmin)
        xmax = max(overx+self.over.bbox.xmax, self.base.bbox.xmax)
        ymin = self.base.bbox.ymin
        ymax = -overy + self.over.bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)


def place_under(base: Mnode, under: Mnode, font: MathFont, emscale: float) -> tuple[float, float]:
    ''' Place node under another one
        Args:
            base: Base node
            under: Node to draw under the base
            font: Font
            emscale: Font scale

        Returns:
            x, y: position for under node
    '''
    # TODO accent parameter used to raise/lower
    x = ((base.bbox.xmax - base.bbox.xmin) - (under.bbox.xmax-under.bbox.xmin)) / 2
    y = -base.bbox.ymin + font.math.consts.underbarVerticalGap*emscale
    y += (under.bbox.ymax)
    return x, y


class Munder(Mnode):
    ''' Under bar '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        kwargs = copy(kwargs)
        assert len(self.element) == 2
        self.base = makenode(self.element[0], size, parent=self, **kwargs)
        kwargs['sub'] = True
        fsize = max(self.size * self.font.math.consts.scriptPercentScaleDown / 100,
                    self.font.basesize*MINFONTSIZE)

        kwargs['width'] = self.base.bbox.xmax - self.base.bbox.xmin
        self.under = makenode(self.element[1], fsize, parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        underx, undery = place_under(self.base, self.under, self.font, self.emscale)

        basex = 0.
        if underx < 0:
            basex = -underx
            underx = 0

        self.nodes.append(self.base)
        self.nodexy.append((basex, 0))
        self.nodes.append(self.under)
        self.nodexy.append((underx, undery))

        xmin = min(underx, self.base.bbox.xmin)
        xmax = max(underx+self.under.bbox.xmax, self.base.bbox.xmax)
        ymin = -undery + self.under.bbox.ymin
        ymax = self.base.bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Munderover(Mnode):
    ''' Under bar and over bar '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        kwargs = copy(kwargs)
        assert len(self.element) == 3
        self.base = makenode(self.element[0], size, parent=self, **kwargs)
        kwargs['sub'] = True
        fsize = max(self.size * self.font.math.consts.scriptPercentScaleDown / 100,
                    self.font.basesize*MINFONTSIZE)

        kwargs['width'] = self.base.bbox.xmax - self.base.bbox.xmin
        self.under = makenode(self.element[1], fsize, parent=self, **kwargs)
        self.over = makenode(self.element[2], fsize, parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes = []
        overx, overy = place_over(self.base, self.over, self.font, self.emscale)
        underx, undery = place_under(self.base, self.under, self.font, self.emscale)

        basex = 0.
        if overx < 0 or underx < 0:
            basex = max(-overx, -underx)
            overx, underx = overx - min(overx, underx), underx - min(overx, underx)

        self.nodes.append(self.base)
        self.nodexy.append((basex, 0))
        self.nodes.append(self.over)
        self.nodes.append(self.under)
        self.nodexy.append((overx, overy))
        self.nodexy.append((underx, undery))
        xmin = min(underx, overx, basex+self.base.bbox.xmin)
        xmax = max(underx+self.under.bbox.xmax, overx+self.over.bbox.xmax, basex+self.base.bbox.xmax)
        ymin = -undery + self.under.bbox.ymin
        ymax = -overy + self.over.bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Mfrac(Mnode):
    ''' Fraction node '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        if kwargs.get('sup') or kwargs.get('sub'):
            # Shrink font another step when frac is in super/subscript
            size *= parent.font.math.consts.scriptPercentScaleDown/100

        super().__init__(element, size, parent, **kwargs)
        assert len(self.element) == 2
        self.size = size
        self.numerator = makenode(self.element[0], size, parent=self, **kwargs)
        self.denominator = makenode(self.element[1], size, parent=self, **kwargs)
        self._setup(**kwargs)
        # TODO: bevelled attribute for x/y fractions with slanty bar

    def _setup(self, **kwargs) -> None:
        ynum = -self.font.math.consts.fractionNumeratorShiftUp * self.emscale
        ydenom = + self.font.math.consts.fractionDenominatorShiftDown * self.emscale
        denombox = self.denominator.bbox
        numbox = self.numerator.bbox

        if ynum + numbox.ymin < 0:
            ynum += numbox.ymin + self.font.math.consts.fractionNumeratorGapMin * self.emscale
        if ydenom - denombox.ymax < 0:
            ydenom -= ydenom - denombox.ymax + self.font.math.consts.fractionDenominatorGapMin * self.emscale

        x = 0.
        if isinstance(self.leftsibling(), Mfrac):
            # Shift a bit so adjacent fracs don't run together
            x += self.size/4
            # COULD DO: adjust denominator ydenom to match the sibling
            # and/or adjust sibling's denominator

        width = max(numbox.xmax, denombox.xmax)
        xnum = x + (width - (numbox.xmax - numbox.xmin))/2
        xdenom = x + (width - (denombox.xmax - denombox.xmin))/2
        self.nodes.append(self.numerator)
        self.nodes.append(self.denominator)
        self.nodexy.append((xnum, ynum))
        self.nodexy.append((xdenom, ydenom))

        linethick = self.font.math.consts.fractionRuleThickness*self.emscale
        if 'linethickness' in self.element.attrib:
            lt = self.element.attrib['linethickness']
            try:
                linethick = getdimension(lt, self.emscale)
            except ValueError:
                linethick = {'thin': linethick * .5,
                             'thick': linethick * 2}.get(lt, linethick)

        bary = -self.font.math.consts.axisHeight*self.emscale
        self.nodes.append(MHLine(width, linethick, **kwargs))
        self.nodexy.append((x, bary))

        # Calculate/cache bounding box
        xmin = 0
        xmax = x + max(numbox.xmax, denombox.xmax)
        ymin = (-ydenom) + denombox.ymin
        ymax = (-ynum) + numbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Mroot(Mnode):
    ''' Nth root '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        self.base = makenode(element[0], size, parent=self, **kwargs)
        dsize = self.size * self.font.math.consts.scriptPercentScaleDown/100
        self.degree: Optional[Mnode]
        self.degree = makenode(element[1], dsize, parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        height = self.base.bbox.ymax - self.base.bbox.ymin

        # Get the right glyph to fit the contents
        rglyph = self.font.math.variant(self.font.glyphindex('√'), height/self.emscale, vert=True)
        rootnode = MGlyph(rglyph, '√', self.size, self.emscale, **kwargs)

        # Shift radical up/down to ensure minimum gap between top of text and overbar
        # Keep contents at the same height.
        rtop = self.base.bbox.ymax + self.font.math.consts.radicalVerticalGap * self.emscale + \
               self.font.math.consts.radicalRuleThickness * self.emscale
        if ((self.base.bbox.ymin < rglyph.path.bbox.ymin*self.emscale) or
            (rglyph.path.bbox.ymax*self.emscale < self.base.bbox.ymax + self.font.math.consts.radicalVerticalGap * self.emscale)):
            yrad = -(rtop - rglyph.path.bbox.ymax * self.emscale)
        else:
            yrad = 0

        ytop = yrad - rglyph.path.bbox.ymax * self.emscale

        # If the root has a degree, draw it next as it
        # determines radical x position
        self.nodes = []
        x = 0
        if self.degree:
            x += self.font.math.consts.radicalKernBeforeDegree * self.emscale
            ydeg = ytop * self.font.math.consts.radicalDegreeBottomRaisePercent/100
            self.nodes.append(self.degree)
            self.nodexy.append((x, ydeg))
            x += self.degree.bbox.xmax
            x += self.font.math.consts.radicalKernAfterDegree * self.emscale

        self.nodes.append(rootnode)
        self.nodexy.append((x, yrad))
        x += rootnode.bbox.xmax
        self.nodes.append(self.base)
        self.nodexy.append((x, 0))
        width = self.base.bbox.xmax - self.base.bbox.xmin

        lastg = self.base.lastglyph()
        if lastg:
            italicx = self.font.math.italicsCorrection.getvalue(lastg.index)
            if italicx:
                width += italicx * self.emscale

        self.nodes.append(MHLine(width, self.font.math.consts.radicalRuleThickness * self.emscale, **kwargs))
        self.nodexy.append((x, yrad-rglyph.path.bbox.ymax * self.emscale))
        xmin = rglyph.path.bbox.xmin * self.emscale
        xmax = x + width
        ymin = min(-yrad + rglyph.path.bbox.ymin * self.emscale, self.base.bbox.ymin)
        if self.degree:
            ymax = max(-yrad + rglyph.path.bbox.ymax * self.emscale, -ydeg+self.degree.bbox.ymax)
        else:
            ymax = -yrad + rglyph.path.bbox.ymax * self.emscale
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Msqrt(Mroot):
    ''' Square root '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        Mnode.__init__(self, element, size, parent, **kwargs)
        if len(self.element) > 1:
            row = ET.Element('mrow')
            row.extend(list(self.element))
            self.base = makenode(row, size, parent=self, **kwargs)
        else:
            self.base = makenode(self.element[0], size, parent=self, **kwargs)
        self.degree = None
        self._setup(**kwargs)


class Mspace(Mnode):
    ''' Blank space '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        self.width = getspaceems(element.attrib.get('width', '0'))  * self.emscale * self.font.info.layout.unitsperem
        self.height = getspaceems(element.attrib.get('height', '0'))  * self.emscale * self.font.info.layout.unitsperem
        self.depth = getspaceems(element.attrib.get('depth', '0'))  * self.emscale * self.font.info.layout.unitsperem
        self._setup(**kwargs)
        
    def _setup(self, **kwargs) -> None:
        self.bbox = BBox(0, self.width, -self.depth, self.height)


class Mpadded(Mrow):
    ''' Mpadded element - Mrow with extra whitespace '''
    def _setup(self, **kwargs):
        super()._setup(**kwargs)
        width = self.element.attrib.get('width', None)
        lspace = self.element.attrib.get('lspace', 0)
        height = self.element.attrib.get('height', None)
        depth = self.element.attrib.get('depth', None)
        xmin, xmax, ymin, ymax = self.bbox

        def adjust(valstr: str, param: float) -> float:
            if valstr.startswith('+') or valstr.startswith('-'):
                sign = width[0]
                if sign == '+':
                    param += getdimension(valstr[1:], self.emscale)
                else:
                    param -= getdimension(valstr[1:], self.emscale)
            elif valstr.endswith('%'):
                param *= float(valstr[:-1])/100
            else:
                param = getdimension(valstr, self.emscale)
            return param

        if width:
            xmax = adjust(width, xmax)
        if height:
            ymax = adjust(height, ymax)
        if depth:
            ymin = -adjust(depth, ymin)
        if lspace:
            xshift = adjust(lspace, 0)
            for i, (x, y) in range(len(self.nodexy)):
                self.nodexy[i] = (x + xshift, y)
        self.bbox = BBox(0, xmax, ymin, ymax)


class Mphantom(Mrow):
    ''' Phantom element. Takes up space but not drawn. '''
    def _setup(self, **kwargs) -> None:
        kwargs['phantom'] = True
        super()._setup(**kwargs)


class Mtable(Mnode):
    ''' Table node '''
    def __init__(self, element: ET.Element, size: float, parent: Mnode, **kwargs):
        super().__init__(element, size, parent, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        kwargs = copy(kwargs)
        rowspace = getdimension('0.2em', self.emscale)
        colspace = getdimension('0.2em', self.emscale)

        # Build node objects from table cells
        rows = []
        for rowelm in self.element:
            assert rowelm.tag == 'mtr'
            cells = []
            for cellelm in rowelm:
                assert cellelm.tag == 'mtd'
                cells.append(makenode(cellelm, self.size, parent=self, **kwargs))
            rows.append(cells)

        # Compute size of each cell to size rows and columns
        rowheights = []  # Maximum height ABOVE baseline
        rowdepths = []   # Maximum distanve BELOW baseline
        for row in rows:
            rowheights.append(max([cell.bbox.ymax for cell in row]))
            rowdepths.append(min([cell.bbox.ymin for cell in row]))

        colwidths = []
        for col in [list(i) for i in zip(*rows)]:  # transposed
            colwidths.append(max([cell.bbox.xmax - cell.bbox.xmin for cell in col]))

        if self.element.attrib.get('equalrows') == 'true':
            rowheights = [max(rowheights)] * len(rows)
            rowdepths = [min(rowdepths)] * len(rows)
        if self.element.attrib.get('equalcolumns') == 'true':
            colwidths = [max(colwidths)] * len(colwidths)

        # Make Baseline of the table half the height
        # Compute baselines to each row
        totheight = sum(rowheights) - sum(rowdepths) + rowspace*(len(rows)-1)
        width = sum(colwidths) + colspace*len(colwidths)
        ytop = -totheight/2
        baselines = []
        y = ytop
        for h, d in zip(rowheights, rowdepths):
            baselines.append(y + h)
            y += h - d + rowspace

        for r, row in enumerate(rows):
            x = colspace/2
            for c, cell in enumerate(row):
                self.nodes.append(cell)
                self.nodexy.append((x, baselines[r]))
                x += colwidths[c] + colspace

        ymin = min([cell.bbox.ymin-baselines[-1] for cell in rows[-1]])
        ymax = max([-baselines[0]+cell.bbox.ymax for cell in rows[0]])
        self.bbox = BBox(0, width, ymin, ymax)



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name = 'ziamath',
    version = '0.2',
    description = 'Render MathML and LaTeX Math to SVG without Latex installation',
    author = 'Collin J. Delker',
    author_email = 'ziaplot@collindelker.com',
    url = 'https://ziamath.readthedocs.io/',
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        'Source': 'https://bitbucket.org/cdelker/ziamath',
    },
    python_requires='>=3.8',
    packages=setuptools.find_packages(),
    package_data = {'ziamath': ['py.typed'],
                    'ziamath.fonts': ['STIXTwoMath-Regular.ttf']},
    zip_safe=False,
    keywords = ['MathML', 'LaTeX', 'math', 'font', 'truetype', 'opentype', 'svg'],
    install_requires=['ziafont'],
    extras_require={
        'latex':  ['latex2mathml'],
    },
    classifiers = [
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering :: Mathematics',
    ],
)



''' Read Math Table from Open Type Font file

    Reference: https://docs.microsoft.com/en-us/typography/opentype/spec/math
'''

from __future__ import annotations
from typing import Union, Sequence, Optional, TYPE_CHECKING
from collections import namedtuple
from dataclasses import dataclass

from ziafont.gpos import Coverage
from ziafont.fontread import FontReader
from ziafont import glyph
from ziafont.fonttypes import Xform, BBox, GlyphComp

if TYPE_CHECKING:
    from .mathfont import MathFont


GlyphType = Union[glyph.SimpleGlyph, glyph.CompoundGlyph]    
    
MathKernInfoRecord = namedtuple(
    'MathKernInfoRecord', ['topright', 'topleft', 'bottomright', 'bottomleft'])
GlyphPartRecord = namedtuple(
    'GlyphPartRecord', ['glyphId', 'startConnectorLength', 'endConnectorLength',
                        'fullAdvance', 'partFlags'])


def read_valuerecord(fontfile: FontReader) -> int:
    ''' Read a Math Value Record. deviceOffset is ignored. '''
    value = fontfile.readint16()
    fontfile.readint16()  # ignore
    return value


@dataclass
class MathConstants:
    ''' Data from the Math Constants Table '''
    scriptPercentScaleDown: int
    scriptScriptPercentScaleDown: int
    delimitedSubFormulaMinHeight: int
    displayOperatorMinHeight: int
    mathLeading: int
    axisHeight: int
    accentBaseHeight: int
    flattenedAccentBaseHeight: int
    subscriptShiftDown: int
    subscriptTopMax: int
    subscriptBaselineDropMin: int
    superscriptShiftUp: int
    superscriptShiftUpCramped: int
    superscriptBottomMin: int
    superscriptBaselineDropMax: int
    subSuperscriptGapMin: int
    superscriptBottomMaxWithSubscript: int
    spaceAfterScript: int
    upperLimitGapMin: int
    upperLimitBaselineRiseMin: int
    lowerLimitGapMin: int
    lowerLimitBaselineDropMin: int
    stackTopShiftUp: int
    stackTopDisplayStyleShiftUp: int
    stackBottomShiftDown: int
    stackBottomDisplayStyleShiftDown: int
    stackGapMin: int
    stackDisplayStyleGapMin: int
    stretchStackTopShiftUp: int
    stretchStackBottomShiftDown: int
    stretchStackGapAboveMin: int
    stretchStackGapBelowMin: int
    fractionNumeratorShiftUp: int
    fractionNumeratorDisplayStyleShiftUp: int
    fractionDenominatorShiftDown: int
    fractionDenominatorDisplayStyleShiftDown: int
    fractionNumeratorGapMin: int
    fractionNumDisplayStyleGapMin: int
    fractionRuleThickness: int
    fractionDenominatorGapMin: int
    fractionDenomDisplayStyleGapMin: int
    skewedFractionHorizontalGap: int
    skewedFractionVerticalGap: int
    overbarVerticalGap: int
    overbarRuleThickness: int
    overbarExtraAscender: int
    underbarVerticalGap: int
    underbarRuleThickness: int
    underbarExtraDescender: int
    radicalVerticalGap: int
    radicalDisplayStyleVerticalGap: int
    radicalRuleThickness: int
    radicalExtraAscender: int
    radicalKernBeforeDegree: int
    radicalKernAfterDegree: int
    radicalDegreeBottomRaisePercent: int


class MathTable:
    ''' OpenType Font MATH Table, provides font-specific kerning adjustments and
        glyph variants for math type setting.

        Args:
            font: Math Font
    '''
    def __init__(self, font: 'MathFont'):
        self.font = font
        self.ofst = font.tables['MATH'].offset
        self.fontfile = self.font.fontfile
        self.fontfile.seek(self.ofst)
        self.vermajor = self.fontfile.readuint16()
        self.verminor = self.fontfile.readuint16()
        assert self.vermajor == 1
        constsofst = self.fontfile.readuint16()
        glyphofst = self.fontfile.readuint16()
        variantsofst = self.fontfile.readuint16()
        self._readconsts(constsofst)
        self._readglyphinfo(glyphofst)
        self._readvariants(variantsofst)

    def _readconsts(self, ofst: int) -> None:
        ''' Read math constants table '''
        self.fontfile.seek(self.ofst + ofst)
        self.consts = MathConstants(
            scriptPercentScaleDown=self.fontfile.readint16(),
            scriptScriptPercentScaleDown=self.fontfile.readint16(),
            delimitedSubFormulaMinHeight=self.fontfile.readuint16(),
            displayOperatorMinHeight=self.fontfile.readuint16(),
            mathLeading=read_valuerecord(self.fontfile),
            axisHeight=read_valuerecord(self.fontfile),
            accentBaseHeight=read_valuerecord(self.fontfile),
            flattenedAccentBaseHeight=read_valuerecord(self.fontfile),
            subscriptShiftDown=read_valuerecord(self.fontfile),
            subscriptTopMax=read_valuerecord(self.fontfile),
            subscriptBaselineDropMin=read_valuerecord(self.fontfile),
            superscriptShiftUp=read_valuerecord(self.fontfile),
            superscriptShiftUpCramped=read_valuerecord(self.fontfile),
            superscriptBottomMin=read_valuerecord(self.fontfile),
            superscriptBaselineDropMax=read_valuerecord(self.fontfile),
            subSuperscriptGapMin=read_valuerecord(self.fontfile),
            superscriptBottomMaxWithSubscript=read_valuerecord(self.fontfile),
            spaceAfterScript=read_valuerecord(self.fontfile),
            upperLimitGapMin=read_valuerecord(self.fontfile),
            upperLimitBaselineRiseMin=read_valuerecord(self.fontfile),
            lowerLimitGapMin=read_valuerecord(self.fontfile),
            lowerLimitBaselineDropMin=read_valuerecord(self.fontfile),
            stackTopShiftUp=read_valuerecord(self.fontfile),
            stackTopDisplayStyleShiftUp=read_valuerecord(self.fontfile),
            stackBottomShiftDown=read_valuerecord(self.fontfile),
            stackBottomDisplayStyleShiftDown=read_valuerecord(self.fontfile),
            stackGapMin=read_valuerecord(self.fontfile),
            stackDisplayStyleGapMin=read_valuerecord(self.fontfile),
            stretchStackTopShiftUp=read_valuerecord(self.fontfile),
            stretchStackBottomShiftDown=read_valuerecord(self.fontfile),
            stretchStackGapAboveMin=read_valuerecord(self.fontfile),
            stretchStackGapBelowMin=read_valuerecord(self.fontfile),
            fractionNumeratorShiftUp=read_valuerecord(self.fontfile),
            fractionNumeratorDisplayStyleShiftUp=read_valuerecord(self.fontfile),
            fractionDenominatorShiftDown=read_valuerecord(self.fontfile),
            fractionDenominatorDisplayStyleShiftDown=read_valuerecord(self.fontfile),
            fractionNumeratorGapMin=read_valuerecord(self.fontfile),
            fractionNumDisplayStyleGapMin=read_valuerecord(self.fontfile),
            fractionRuleThickness=read_valuerecord(self.fontfile),
            fractionDenominatorGapMin=read_valuerecord(self.fontfile),
            fractionDenomDisplayStyleGapMin=read_valuerecord(self.fontfile),
            skewedFractionHorizontalGap=read_valuerecord(self.fontfile),
            skewedFractionVerticalGap=read_valuerecord(self.fontfile),
            overbarVerticalGap=read_valuerecord(self.fontfile),
            overbarRuleThickness=read_valuerecord(self.fontfile),
            overbarExtraAscender=read_valuerecord(self.fontfile),
            underbarVerticalGap=read_valuerecord(self.fontfile),
            underbarRuleThickness=read_valuerecord(self.fontfile),
            underbarExtraDescender=read_valuerecord(self.fontfile),
            radicalVerticalGap=read_valuerecord(self.fontfile),
            radicalDisplayStyleVerticalGap=read_valuerecord(self.fontfile),
            radicalRuleThickness=read_valuerecord(self.fontfile),
            radicalExtraAscender=read_valuerecord(self.fontfile),
            radicalKernBeforeDegree=read_valuerecord(self.fontfile),
            radicalKernAfterDegree=read_valuerecord(self.fontfile),
            radicalDegreeBottomRaisePercent=self.fontfile.readint16())

    def _readglyphinfo(self, ofst: int) -> None:
        ''' Read glyph info table '''
        self.fontfile.seek(self.ofst + ofst)
        # Read table offsets
        italics = self.fontfile.readuint16()
        topaccent = self.fontfile.readuint16()
        extendshape = self.fontfile.readuint16()
        kernofst = self.fontfile.readuint16()

        # Italics table
        self.fontfile.seek(self.ofst + ofst + italics)
        covofst = self.fontfile.readuint16()
        cnt = self.fontfile.readuint16()
        italicscorrections = []
        for i in range(cnt):
            italicscorrections.append(read_valuerecord(self.fontfile))
        italicscoverage = Coverage(self.ofst+ofst+italics+covofst, self.fontfile)

        # Top Accent Attachment Table
        self.fontfile.seek(self.ofst + ofst + topaccent)
        covofst = self.fontfile.readuint16()
        cnt = self.fontfile.readuint16()
        accents = []
        for i in range(cnt):
            accents.append(read_valuerecord(self.fontfile))
        accentcoverage = Coverage(self.ofst+ofst+topaccent+covofst, self.fontfile)

        # Extended Shape Coverage
        extshapes = Coverage(self.ofst+ofst+extendshape, self.fontfile)

        # Kern Info
        kerninfo = MathKernInfoTable(self.ofst+ofst+kernofst, self.fontfile)

        self.italicsCorrection = MathSubTable(italicscorrections, italicscoverage)
        self.topAccentAttachment = MathSubTable(accents, accentcoverage)
        self._extendedShapeCoverage = extshapes
        self.kernInfo = kerninfo

    def _readvariants(self, ofst: int) -> None:
        ''' Read the variants table '''
        self.fontfile.seek(self.ofst + ofst)
        minoverlap = self.fontfile.readuint16()
        vertcovofst = self.fontfile.readuint16()
        horzcovofst = self.fontfile.readuint16()
        vertcount = self.fontfile.readuint16()
        horzcount = self.fontfile.readuint16()

        vertConstruction = []
        for i in range(vertcount):
            vofst = self.fontfile.readuint16()
            vertConstruction.append(MathConstructionTable(self.ofst+ofst+vofst, self.font, vert=True))
        horzConstruction = []
        for i in range(horzcount):
            hofst = self.fontfile.readuint16()
            horzConstruction.append(MathConstructionTable(self.ofst+ofst+hofst, self.font, vert=False))

        vertcoverage = Coverage(self.ofst + ofst + vertcovofst, self.fontfile)
        horzcoverage = Coverage(self.ofst + ofst + horzcovofst, self.fontfile)
        self._variantsvert = MathVariants(vertcoverage, vertConstruction, minoverlap, self.font, vert=True)
        self._variantshorz = MathVariants(horzcoverage, horzConstruction, minoverlap, self.font, vert=False)

    def kernsuper(self, glyph1: GlyphType, glyph2: GlyphType) -> tuple[int, int]:
        ''' Calculate superscript kerning between the two glyphs

            Args:
                glyph1: Last glyph in base
                glyph2: First glyph in superscript

            Returns:
                kern: Kerning shift to apply in x direction
                shift: Upward shift of superscript with respect to baseline
        '''
        g1 = self.kernInfo.glyph(glyph1.index)
        g2 = self.kernInfo.glyph(glyph2.index)

        # Extended shape, need to raise superscript up, but only if we're using
        # a variant
        if self._extendedShapeCoverage.covidx(glyph1.index) is not None:
            shiftup = glyph1.path.bbox.ymax - self.consts.superscriptShiftUp/2
        else:
            shiftup = self.consts.superscriptShiftUp

        h1 = shiftup + glyph2.path.bbox.ymin * self.consts.scriptPercentScaleDown/100
        h2 = glyph1.path.bbox.ymax - shiftup
        k1 = k2 = 0
        if g1:
            k1 += g1.topright.getkern(h1)
            k2 += g1.topright.getkern(h2)
        if g2:
            k1 += g2.bottomleft.getkern(h1)
            k2 += g2.bottomleft.getkern(h2)
        return min(k1, k2), shiftup

    def kernsub(self, glyph1: GlyphType, glyph2: GlyphType) -> tuple[int, int]:
        ''' Calculate subscript kerning

            Args:
                glyph1: Last glyph in base
                glyph2: First glyph in subscript

            Returns:
                kern: Kerning shift to apply in x direction
                shift: Downward shift of subscript with respect to baseline
        '''
        g1 = self.kernInfo.glyph(glyph1.index)
        g2 = self.kernInfo.glyph(glyph2.index)

        # Correction heights
        shiftdn = self.consts.subscriptShiftDown - glyph1.path.bbox.ymin
        h1 = -shiftdn + glyph2.path.bbox.ymax * self.consts.scriptPercentScaleDown/100
        h2 = glyph1.path.bbox.ymin + shiftdn
        k1 = k2 = 0
        if g1:
            k1 += g1.bottomright.getkern(h1)
            k2 += g1.bottomright.getkern(h2)
        if g2:
            k1 += g2.topleft.getkern(h1)
            k2 += g2.topleft.getkern(h2)
        return min(k1, k2), shiftdn

    def variant(self, glyphid: int, height: float, vert: bool=True) -> GlyphType:
        ''' Get a height variant for the glyph

            glyphid: Glyph index
            height: Desired height of glyph in font units
            vert: Vertical variant or horizontal variant
        '''
        if vert:
            variant = self._variantsvert.getvariant(glyphid, height)
        else:
            variant = self._variantshorz.getvariant(glyphid, height)
        return variant

    def isextended(self, glyphid: int) -> bool:
        ''' Determine if glyph is an extended shape (has stretchy variants) '''
        return self._extendedShapeCoverage.covidx(glyphid) is not None


class MathConstructionTable:
    ''' Math Construction Table, listing size variants for a glyph

        Args:
            ofst: Byte offset to table in font file
            font: Font
            vert: Vertical or horizontal variant
    '''
    def __init__(self, ofst: int, font: 'MathFont', vert: bool=True):
        fileptr = font.fontfile.tell()
        font.fontfile.seek(ofst)
        assemblyofst = font.fontfile.readuint16()
        varcount = font.fontfile.readuint16()
        self.variants = {}
        for i in range(varcount):
            varglyph = font.fontfile.readuint16()
            advmeas = font.fontfile.readuint16()
            self.variants[advmeas] = varglyph

        self.assembly: Optional[MathAssembly]
        if assemblyofst:
            self.assembly = MathAssembly(ofst+assemblyofst, font, vert=vert)
        else:
            self.assembly = None
        font.fontfile.seek(fileptr)


class MathAssembly:
    ''' Math assembly table, for combining several glyphs to extend the size
        beyond the font built-in glyphs (for example, making a really big
        curly brace)

        Args:
            ofst: Byte offset into font file
            font: Font
            vert: Vertical or horizontal variant
    '''
    def __init__(self, ofst: int, font: 'MathFont', vert: bool=True):
        self.vert = vert
        self.font = font
        font.fontfile.seek(ofst)
        self.italicscorrection = read_valuerecord(font.fontfile)
        partcnt = font.fontfile.readuint16()
        self.parts = []
        for i in range(partcnt):
            self.parts.append(GlyphPartRecord(
                font.fontfile.readuint16(),
                font.fontfile.readuint16(),
                font.fontfile.readuint16(),
                font.fontfile.readuint16(),
                font.fontfile.readuint16()))

    def assemble(self, reqsize: float, minoverlap: float) -> glyph.CompoundGlyph:
        ''' Build glyph assembly, combining parts to create any required size

            Args:
                reqsize: Desired glyph size
                minoverlap: Minimum overlap from variants table
        '''
        size = 0.
        N = 0

        # Determine number of extender parts needed.
        while size < reqsize:
            # Min overlap give largest size with this many extenders

            testparts = []
            for part in self.parts:
                if part.partFlags:
                    testparts.extend([part]*N)
                else:
                    testparts.append(part)

            y = 0.
            for i, part in enumerate(testparts):
                if i > 0:
                    y -= minoverlap
                y += part.fullAdvance
            size = y
            N += 1

        # Decrease overlap since full extenders make it too tall
        dy = (size - reqsize) / (len(testparts)-1)

        # Build transforms for compound glyph
        xforms = []
        if self.vert:
            y = -reqsize/2 + self.font.math.consts.axisHeight
            ymin = y
        else:
            y = 0.
        for i, part in enumerate(testparts):
            if i > 0:
                y -= minoverlap + dy
            if self.vert:
                xforms.append(Xform(1, 0, 0, 1, 0, y, False))
            else:
                xforms.append(Xform(1, 0, 0, 1, y, 0, False))
            y += part.fullAdvance
        size = y

        # Put together the CompoundGlyph
        glyphs = [self.font.glyph_fromid(part.glyphId) for part in testparts]
        
        # Make a unique ID, negative so it can't clash with other glyphs
        glyfid = -(testparts[0].glyphId + int(reqsize) << 16)

        if self.vert:
            xmin = min([g.path.bbox.xmin for g in glyphs])
            xmax = max([g.path.bbox.xmax for g in glyphs])
            ymin = min([ymin+g.path.bbox.ymin for g in glyphs])
            ymax = size
        else:
            xmin = min([g.path.bbox.xmin for g in glyphs])
            xmax = size
            ymin = min([g.path.bbox.ymin for g in glyphs])
            ymax = max([g.path.bbox.ymax for g in glyphs])

        bbox = BBox(xmin, xmax, ymin, ymax)
        gc = GlyphComp(glyphs, xforms, bbox)
        glf = glyph.CompoundGlyph(glyfid, gc, self.font)
        glf.advance = lambda x=bbox.xmax: x  # type: ignore
        return glf


class MathVariants:
    ''' Math Variants Table. Provides different glyphs for different size symbols,
        such as curly braces and square roots.
    '''
    def __init__(self, coverage: Coverage, construction: Sequence[MathConstructionTable],
                 minoverlap: float, font: 'MathFont', vert: bool):
        self.coverage = coverage
        self.construction = construction
        self.minoverlap = minoverlap
        self.font = font
        self.vert = vert

    def getvariant(self, glyphid: int, size: float) -> GlyphType:
        ''' Get the proper size variant for the glyphid '''
        covidx = self.coverage.covidx(glyphid)
        if covidx is None:
            # Not covered by a variant
            return self.font.glyph_fromid(glyphid)

        construction = self.construction[covidx]
        variants = construction.variants
        glf: GlyphType
        try:
            variantid = variants[min(k for k in variants if k >= size)]

        except ValueError:
            # size is bigger than all variants. Use assembly table.
            if construction.assembly:
                glf = construction.assembly.assemble(size, self.minoverlap)
            else:
                glf = self.font.glyph_fromid(variants[max(variants.keys())])
        else:
            glf = self.font.glyph_fromid(variantid)

        return glf


class MathKernTable:
    ''' Math Kerning Table, for adjusting sub/superscripts

        Args:
            ofst: Byte offset into font file
            fontfile: font file
    '''
    def __init__(self, ofst: int, fontfile: FontReader):
        fileptr = fontfile.tell()
        fontfile.seek(ofst)
        heightcnt = fontfile.readuint16()
        self.heights = []
        for i in range(heightcnt):
            self.heights.append(read_valuerecord(fontfile))
        self.kernvalues = []
        for i in range(heightcnt+1):
            self.kernvalues.append(read_valuerecord(fontfile))
        fontfile.seek(fileptr)

    def getkern(self, height: float) -> int:
        ''' Get kerning for this height '''
        if len(self.heights) == 0:
            return self.kernvalues[0]

        for i, h in enumerate(self.heights):
            if h > height:
                break
        else:
            i += 1
        return self.kernvalues[i]


class ZeroKern:
    ''' A kerning table with 0 kerning '''
    def getkern(self, height):
        return 0


class MathKernInfoTable:
    ''' Math Kerning Info Table, listing of MathKernTables

        Args:
            ofst: Byte offset into font file
            fontfile: Font File
    '''
    def __init__(self, ofst: int, fontfile: FontReader):
        fontfile.seek(ofst)
        covofst = fontfile.readuint16()
        cnt = fontfile.readuint16()
        self.kerninfo = []
        for i in range(cnt):
            tr = fontfile.readuint16()
            tl = fontfile.readuint16()
            br = fontfile.readuint16()
            bl = fontfile.readuint16()
            self.kerninfo.append(MathKernInfoRecord(
                MathKernTable(tr+ofst, fontfile) if tr else ZeroKern(),
                MathKernTable(tl+ofst, fontfile) if tl else ZeroKern(),
                MathKernTable(br+ofst, fontfile) if br else ZeroKern(),
                MathKernTable(bl+ofst, fontfile) if bl else ZeroKern()))
        self.coverage = Coverage(ofst+covofst, fontfile)

    def glyph(self, glyphid: int) -> Optional[MathKernInfoRecord]:
        ''' Get kerning info record for this glyph '''
        idx = self.coverage.covidx(glyphid)
        if idx is not None:
            return self.kerninfo[idx]
        return None


class MathSubTable:
    ''' Math Sub Table - generic list of values associated with glyphs '''
    def __init__(self, values: Sequence[float], coverage: Coverage):
        self.values = values
        self.coverage = coverage

    def getvalue(self, glyphid: int) -> Optional[float]:
        ''' Get a value from the table for the glyph '''
        idx = self.coverage.covidx(glyphid)
        if idx is not None:
            return self.values[idx]
        return None
